@@ -2,7 +2,7 @@
 /*
  *  Authors: Jeffrey Stedfast <fejj@ximian.com>
  *
- *  Copyright 2001 Ximian, Inc. (www.ximian.com)
+ *  Copyright 2001-2004 Ximian, Inc. (www.ximian.com)
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -77,7 +77,12 @@ g_mime_disposition_new (const char *disposition)
 	}
 	
 	/* get content disposition part */
-	for (inptr = disposition; *inptr && *inptr != ';'; inptr++); /* find ; or \0 */
+	
+	/* find ; or \0 */
+	inptr = disposition;
+	while (*inptr && *inptr != ';')
+		inptr++;
+	
 	value = g_strndup (disposition, (unsigned) (inptr - disposition));
 	g_strstrip (value);
 	
